from typing import List, Tuple

from ..constants import MIN_ROW, MAX_ROW, MIN_COL, MAX_COL, HAN_ADVANTAGE
from ..base.board import Board
from ..base.camp import Camp
from ..base.formation import Formation
from ..base.piece import PieceType
from ..base.location import Location
from ..base.move import MoveSet
from .game_log import GameLog


class JanggiGame:
    """
    A game of Janggi with a game board, players, and scores.
    With two public methods, this class can help make a single action on the board,
    and it can also get all possible moves from a current state of the board.
    """

    def __init__(self, player: Camp, cho_formation: Formation, han_formation: Formation):
        """
        Initialize Janggi game instance.

        Args:
            player (Camp): Camp that the main player is playing.
            cho_formation (Formation): Formation of camp cho.
            han_formation (Formation): Formation of camp han.
        """
        self.player = player
        self.cho_formation = cho_formation
        self.han_formation = han_formation

        self.turn = Camp.CHO
        self.cho_score = self.han_score = 0.0
        self.log = GameLog(cho_formation, han_formation, player)
        self.board = Board.full_board_from_formations(cho_formation, han_formation, player)
        self.initial_board = self.board.copy()
        self._update_scores()

    def make_action(self, origin: Location, dest: Location) -> Tuple[float, bool]:
        """
        Move a piece from the given origin location to the given destination.

        Args:
            origin (Location): Original location of the piece to be moved.
            dest (Location): Destination of the piece to be moved.

        Raises:
            Exception: When the given action is invalid.

        Returns:
            piece_value (float): An enemy piece's value if it was killed; 0 otherwise.
            game_over (bool): True if the action ends the game; False otherwise.
        """
        # validate the given action
        if not self._validate_action(origin, dest):
            raise Exception("The action cannot be made!")

        # make the action
        piece_removed = self.board.move(origin, dest)

        # detenmine if game's over
        game_over = False
        piece_value = 0
        if piece_removed:
            piece_value = piece_removed.value
            if piece_removed.piece_type == PieceType.GENERAL:
                game_over = True

        # update cho and han's scores
        self._update_scores()

        # switch "turn"
        self.turn = self.turn.opponent

        # record move logs
        self.log.add_move((origin, dest))

        return float(piece_value), game_over

    def get_all_actions(self) -> List[Tuple[Location, Location]]:
        """
        Get list of all possible moves that can be made for the current player.

        Returns:
            List[Tuple[Location, Location]]: List of moves in (origin, dest) format 
              where it means a piece at origin location being moved to dest location.
        """
        possible_actions = []
        piece_locations = self.board.get_piece_locations_for_camp(self.turn)
        for piece_location in piece_locations:
            destinations = self.get_all_destinations(piece_location)
            possible_actions += [(piece_location, dest_location)
                                 for dest_location in destinations]
        return possible_actions

    def get_all_destinations(self, origin: Location) -> List[Location]:
        """
        List all possible locations where a piece at given origin can move to.

        Args:
            origin (Location): Location of the piece to get destionations for.

        Returns:
            List[Location]: List of all possible locations the piece can go to.
        """
        move_sets = self._get_possible_move_sets(origin)
        all_dest = [ms.get_dest(self.board, origin, self.turn)
                    for ms in move_sets]
        return all_dest

    def _update_scores(self):
        """
        Update cho and han's scores by summing up each of their piece's value.
        """
        self.cho_score = self.board.get_score(Camp.CHO)
        self.han_score = self.board.get_score(Camp.HAN) + HAN_ADVANTAGE

    def _get_possible_move_sets(self, origin: Location) -> List[MoveSet]:
        """
        Get all move sets that pertain to a piece at origin location.
        In other words, get list of all moves a piece at origin can make, as MoveSet instances.

        Args:
            origin (Location): Location of the piece in question.

        Raises:
            Exception: When the given input is invalid.

        Returns:
            List[MoveSet]: MoveSets that the piece can possibly perform.
        """
        def _is_out_of_bound(location: Location):
            return (location.row < MIN_ROW or location.row > MAX_ROW or
                    location.col < MIN_COL or location.col > MAX_COL)

        piece = self.board.get(origin.row, origin.col)
        # Invalidate when given locations are out of bound
        if _is_out_of_bound(origin):
            raise Exception(f"The location {origin} is out of bound.")

        # Invalidate when no piece is on "origin" location
        if not piece:
            raise Exception(f"There is not piece on the location {origin}.")

        # Invalidate when the piece does not belong to the current player
        if piece.camp != self.turn:
            raise Exception(f"The piece {piece.piece_type} does not belong to the current player {self.turn}.")

        # Get MoveSets based on piece type
        if piece.piece_type == PieceType.SOLDIER:
            move_sets = piece.get_soldier_move_sets(origin, self.player == self.turn)
        elif piece.piece_type == PieceType.HORSE or piece.piece_type == PieceType.ELEPHANT:
            move_sets = piece.get_jumpy_move_sets()
        elif piece.piece_type == PieceType.CHARIOT or piece.piece_type == PieceType.CANNON:
            move_sets = piece.get_straight_move_sets(origin)
        elif piece.piece_type == PieceType.GENERAL or piece.piece_type == PieceType.GUARD:
            move_sets = piece.get_castle_move_sets(origin, self.player == self.turn)

        # Filter out all the invalid move sets
        move_sets = [ms for ms in move_sets if ms.is_valid(
            self.board, origin, self.turn)]
        return move_sets

    def _validate_action(self, origin: Location, dest: Location) -> bool:
        """
        Check if the given action is valid.

        Args:
            origin (Location): Original location of the piece in question.
            dest (Location): Desired location for the piece to be moved.

        Returns:
            bool: True if action is valid; False otherwise.
        """
        def _is_out_of_bound(location: Location):
            return (location.row < MIN_ROW or location.row > MAX_ROW or
                    location.col < MIN_COL or location.col > MAX_COL)

        origin_piece = self.board.get(origin.row, origin.col)
        dest_piece = self.board.get(dest.row, dest.col)
        # Invalidate when given locations are out of bound
        if _is_out_of_bound(origin) or _is_out_of_bound(dest):
            return False

        # Invalidate when no piece is on "origin" location
        if not origin_piece:
            return False

        # Invalidate when wrong camp
        if origin_piece.camp != self.turn:
            return False

        # Invalidate when ally piece is already on "dest" location
        if dest_piece and dest_piece.camp == origin_piece.camp:
            return False

        # See if destination is in list of all possible destinatins from origin
        if dest not in self.get_all_destinations(origin):
            return False

        # TODO: Invalidate the move makes it enemy's "Janggun"
        return True



from __future__ import annotations
from sqlite3 import Row
from typing import Iterable
from ..constants import MIN_ROW, MAX_ROW, MIN_COL, MAX_COL
from ..proto import log_pb2


class Location:
    """
    Location class that each represents a single location on the board grid.
    """
    def __init__(self, row: int, col: int):
        """
        Initialize location.

        Args:
            row (int): row number in range 0 <= row <= 9.
            col (int): column number in range 0 <= col <= 8.

        Raises:
            Exception: row is out of range.
            Exception: column is out of range.
        """
        self.row = row
        self.col = col

        if self.row < MIN_ROW or self.row > MAX_ROW:
            raise Exception(f"location row is out of range: {self.row}")

        if self.col < MIN_COL or self.col > MAX_COL:
            raise Exception(f"location column is out of range: {self.col}")

    def __str__(self) -> str:
        """Return string representation of location."""
        return f"({self.row},{self.col})"

    def __iter__(self) -> Iterable:
        """Make location iterable so that it can easily be converted into a tuple or list."""
        yield self.row
        yield self.col

    def __eq__(self, other) -> bool:
        """Return True if two location classes have same row and col; False otherwise."""
        return self.row == other.row and self.col == other.col

    @classmethod
    def from_proto(cls, location_proto: log_pb2.Location) -> Location:
        """Convert from proto Location message."""
        return Location(location_proto.row, location_proto.col)

    def to_proto(self) -> log_pb2.Location:
        """Convert to proto Location message."""
        location_proto = log_pb2.Location()
        location_proto.row = self.row
        location_proto.col = self.col
        return location_proto



import logging
import pygame
from pygame.locals import *
from typing import List, Tuple

from ..base.location import Location
from ..constants import MIN_ROW, MAX_ROW, MIN_COL, MAX_COL
from ..game.janggi_game import JanggiGame
from .game_window import (
    GameWindow,
    BOARD_X, BOARD_Y, 
    ROW_GAP, COL_GAP,
)
from .marker import BoardMarker

class MoveSelection:
    def __init__(self, origin: Location, dest: List[Location]):
        self.origin = origin
        self.dest = dest

class GamePlayer:
    """Class used to play the game."""
    def __init__(self, game: JanggiGame):
        self.game = game
        self.window = GameWindow(game.board)
        self.move_selection = None

    def run(self):
        self.window.render()
        pygame.event.clear()
        while True:
            event = pygame.event.wait()
            if event.type == QUIT:
                self.window.close()
                return
            if event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    self.window.close()
                    return
            if event.type == MOUSEBUTTONUP and event.button == 1: # left click
                mousex, mousey = pygame.mouse.get_pos()
                row, col, success = self._get_board_row_col(mousex, mousey)
                if success:
                    logging.info(f"Clicked {self.game.board.get(row, col)}")
                    # Check for moves if currently selecting moves
                    if self.move_selection:
                        if self._check_for_move(row, col):
                            game_over = self._move(row, col)
                            if game_over:
                                self.window.close()
                                return
                        self._clear_selection()
                    # Detect clicks on pieces and set selection if necessary
                    piece = self.game.board.get(row, col)
                    if piece and piece.camp == self.game.turn:
                        self._set_selection(row, col)
                    self.window.render()

    def _check_for_move(self, row: int, col: int):
        for dest_row, dest_col in self.move_selection.dest:
            if row == dest_row and col == dest_col:
                return True
        return False

    def _move(self, row: int, col: int) -> bool:
        assert self.move_selection is not None
        _, game_over = self.game.make_action(
            self.move_selection.origin,
            Location(row, col),
        )
        return game_over

    def _set_selection(self, row: int, col: int):
        dest = self.game.get_all_destinations(Location(row, col))
        self.move_selection = MoveSelection(Location(row, col), dest)
        self.window.board_markers = [
            BoardMarker(
                self.window.display, 
                *(self.window.get_board_xy(dest_row, dest_col)),
            ) for dest_row, dest_col in dest
        ]
    
    def _clear_selection(self):
        self.move_selection = []
        self.window.board_markers = []

    def _get_board_row_col(self, x: int, y: int) -> Tuple[int, int, bool]:
        row = round((y - BOARD_Y - ROW_GAP * 0.5) / ROW_GAP)
        col = round((x - BOARD_X - COL_GAP * 0.5) / COL_GAP)
        if row < MIN_ROW or row > MAX_ROW or col < MIN_COL or col > MAX_COL:
            return -1, -1, False
        return row, col, True



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: janggi/proto/log.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16janggi/proto/log.proto\x12\x06janggi\"$\n\x08Location\x12\x0b\n\x03row\x18\x01 \x01(\x05\x12\x0b\n\x03\x63ol\x18\x02 \x01(\x05\"H\n\x04Move\x12 \n\x06origin\x18\x01 \x01(\x0b\x32\x10.janggi.Location\x12\x1e\n\x04\x64\x65st\x18\x02 \x01(\x0b\x32\x10.janggi.Location\"\x99\x01\n\x03Log\x12(\n\rcho_formation\x18\x01 \x01(\x0e\x32\x11.janggi.Formation\x12(\n\rhan_formation\x18\x02 \x01(\x0e\x32\x11.janggi.Formation\x12!\n\x0b\x62ottom_camp\x18\x03 \x01(\x0e\x32\x0c.janggi.Camp\x12\x1b\n\x05moves\x18\x04 \x03(\x0b\x32\x0c.janggi.Move*\x18\n\x04\x43\x61mp\x12\x07\n\x03\x43HO\x10\x00\x12\x07\n\x03HAN\x10\x01*Z\n\tFormation\x12\x12\n\x0eINNER_ELEPHANT\x10\x00\x12\x12\n\x0eOUTER_ELEPHANT\x10\x01\x12\x11\n\rLEFT_ELEPHANT\x10\x02\x12\x12\n\x0eRIGHT_ELEPHANT\x10\x03\x62\x06proto3')

_CAMP = DESCRIPTOR.enum_types_by_name['Camp']
Camp = enum_type_wrapper.EnumTypeWrapper(_CAMP)
_FORMATION = DESCRIPTOR.enum_types_by_name['Formation']
Formation = enum_type_wrapper.EnumTypeWrapper(_FORMATION)
CHO = 0
HAN = 1
INNER_ELEPHANT = 0
OUTER_ELEPHANT = 1
LEFT_ELEPHANT = 2
RIGHT_ELEPHANT = 3


_LOCATION = DESCRIPTOR.message_types_by_name['Location']
_MOVE = DESCRIPTOR.message_types_by_name['Move']
_LOG = DESCRIPTOR.message_types_by_name['Log']
Location = _reflection.GeneratedProtocolMessageType('Location', (_message.Message,), {
  'DESCRIPTOR' : _LOCATION,
  '__module__' : 'janggi.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:janggi.Location)
  })
_sym_db.RegisterMessage(Location)

Move = _reflection.GeneratedProtocolMessageType('Move', (_message.Message,), {
  'DESCRIPTOR' : _MOVE,
  '__module__' : 'janggi.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:janggi.Move)
  })
_sym_db.RegisterMessage(Move)

Log = _reflection.GeneratedProtocolMessageType('Log', (_message.Message,), {
  'DESCRIPTOR' : _LOG,
  '__module__' : 'janggi.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:janggi.Log)
  })
_sym_db.RegisterMessage(Log)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _CAMP._serialized_start=302
  _CAMP._serialized_end=326
  _FORMATION._serialized_start=328
  _FORMATION._serialized_end=418
  _LOCATION._serialized_start=34
  _LOCATION._serialized_end=70
  _MOVE._serialized_start=72
  _MOVE._serialized_end=144
  _LOG._serialized_start=147
  _LOG._serialized_end=300
# @@protoc_insertion_point(module_scope)



from __future__ import annotations
import numpy as np
from typing import List, Optional

from ..constants import MIN_ROW, MAX_ROW, MIN_COL, MAX_COL, NUM_ROWS, NUM_COLS
from .piece import Piece, PieceType
from .camp import Camp
from .formation import Formation
from .location import Location


class Board:
    """
    Simple board class used for the game of Janggi. Contains and handles a single 
    10x9 two-dimensional list that contains either a Piece object or None.
    """
    def __init__(self, cho_formation: Formation, han_formation: Formation, bottom_camp: Camp):
        self.cho_formation = cho_formation
        self.han_formation = han_formation
        self.bottom_camp = bottom_camp
        self.__board = np.full((NUM_ROWS, NUM_COLS), None)

    def __str__(self) -> str:
        """Generate colored and structured string representation of the board."""
        print_str = ""
        for row in range(-1, MAX_ROW+1):
            for col in range(-1, MAX_COL+1):
                if row == -1 and col >= 0:
                    print_str += " " + str(col)
                elif row >= 0 and col == -1:
                    print_str += " " + str(row % 10)
                elif row >= 0 and col >= 0:
                    if self.__board[row][col]:
                        print_str += str(self.__board[row][col])
                    else:
                        print_str += "  "
                else:
                    print_str += " "
                print_str += " "
            print_str += "\n"
        return print_str

    @classmethod
    def full_board_from_formations(cls, cho_formation: Formation, han_formation: Formation, player: Camp) -> Board:
        """
        Return Board class instance that represents a full board.

        Args:
            cho_formation (Formation): Formation of Camp Cho.
            han_formation (Formation): Formation of Camp Han.
            player (Camp): Camp that the player is playing. This is used to assign that camp as the bottom camp.

        Returns:
            Board: Full Board class.
        """
        board = cls(cho_formation, han_formation, player)
        cho_half_board = Board._generate_half_board(cho_formation)
        han_half_board = Board._generate_half_board(han_formation)
        cho_half_board.mark_camp(Camp.CHO)
        han_half_board.mark_camp(Camp.HAN)
        if player == Camp.CHO:
            han_half_board.rotate()
        else:
            cho_half_board.rotate()
        board.merge(cho_half_board)
        board.merge(han_half_board)

        return board

    def copy(self) -> Board:
        """
        Return a copied Board class.

        Returns:
            Board: Copied version of the board.
        """
        copied_board = Board(self.cho_formation, self.han_formation, self.bottom_camp)
        copied_board.__board = self.__board.copy()
        return copied_board

    def put(self, row: int, col: int, piece: Piece):
        """
        Put piece into board at the given (row,col) location.
        Used row and col as inputs instead of Location to make it easier to generate
        initial boards in formation.py.

        Args:
            row (int): Row that the given piece that will placed on.
            col (int): Column that the given piece that will be placed on.
            piece (Piece): Piece that will be placed on the board.
        """
        self.__board[row][col] = piece

    def merge(self, board: Board):
        """
        Merge the given board into self.__board by overwriting.

        Args:
            board (Board): Input board that will be merged into self.__board.
        """
        for row in range(MIN_ROW, MAX_ROW+1):
            for col in range(MIN_COL, MAX_COL+1):
                if board.get(row, col):
                    self.__board[row][col] = board.get(row, col)

    def get(self, row: int, col: int) -> Piece:
        """
        Return the piece that is located at the given location of the board.

        Args:
            row (int): Row of piece.
            col (int): Column of piece.

        Returns:
            Piece: Piece located at (row,col) on the board. Can be None.
        """
        return self.__board[row][col]

    def remove(self, row: int, col: int):
        """
        Remove piece at the given location of the board.

        Args:
            row (int): Row of the piece to be removed.
            col (int): Column of the piece to be removed.
        """
        self.__board[row][col] = None

    def move(self, origin: Location, dest: Location) -> Optional[Piece]:
        """
        Move piece from origin to destination and return the piece that was
        originally placed at the given dest.

        Args:
            origin (Location): Original location of the piece being played.
            dest (Location): Destination of the piece being played.
        """
        assert self.get(origin.row, origin.col) is not None
        piece_to_remove = self.get(dest.row, dest.col)
        piece = self.get(origin.row, origin.col)
        self.put(dest.row, dest.col, piece)
        self.remove(origin.row, origin.col)
        return piece_to_remove

    def rotate(self):
        """Rotate the board 180 degrees and update self.__board."""
        new_board = np.full((NUM_ROWS, NUM_COLS), None)
        for row in range(MIN_ROW, MAX_ROW+1):
            for col in range(MIN_COL, MAX_COL+1):
                if self.__board[row][col]:
                    new_board[NUM_ROWS-row-1][NUM_COLS-col-1] = self.__board[row][col]
        self.__board = new_board

    def mark_camp(self, camp: Camp):
        """
        Mark all pieces on the board with the given camp (CHO or HAN).
        Used to mark half-boards when setting up initial boards.

        Args:
            camp (Camp): Camp enum to mark pieces with.
        """
        for row in range(MIN_ROW, MAX_ROW+1):
            for col in range(MIN_COL, MAX_COL+1):
                if self.__board[row][col]:
                    self.__board[row][col].camp = camp

    def get_score(self, camp: Camp) -> int:
        """
        Return score for the player who's playing the given camp.

        Args:
            camp (Camp): Camp of the player whose score will be calculated.

        Returns:
            int: Score of the player who's playing the given camp.
        """
        score = 0
        for row in range(MIN_ROW, MAX_ROW+1):
            for col in range(MIN_COL, MAX_COL+1):
                if self.__board[row][col] and self.__board[row][col].camp == camp:
                    score += self.__board[row][col].value
        return score

    def get_piece_locations(self) -> List[Location]:
        """
        Get locations of all pieces on the board.

        Returns:
            List[Location]: List of all locations of the pieces on the board.
        """
        piece_locations = []
        for row in range(MIN_ROW, MAX_ROW+1):
            for col in range(MIN_COL, MAX_COL+1):
                if self.__board[row][col]:
                    piece_locations.append(Location(row, col))
        return piece_locations

    def get_piece_locations_for_camp(self, camp: Camp) -> List[Location]:
        """
        Get locations of all pieces with the given camp.

        Args:
            camp (Camp): Camp of the pieces to fetch.

        Returns:
            List[Location]: List of all locations of the pieces with the given camp.
        """
        piece_locations = []
        for row in range(MIN_ROW, MAX_ROW+1):
            for col in range(MIN_COL, MAX_COL+1):
                if self.__board[row][col] and self.__board[row][col].camp == camp:
                    piece_locations.append(Location(row, col))
        return piece_locations

    @classmethod
    def _generate_half_board(cls, formation: Formation) -> Board:
        """
        Generate half-board that only contains pieces for a single side.

        Args:
            formation (Formation): Formation of the half board.

        Returns:
            Board: Half-board that contains pieces for only one camp.
        """
        board = cls(Formation.UNDECIDED, Formation.UNDECIDED, Camp.UNDECIDED)
        board.put(6, 0, Piece(PieceType.SOLDIER))
        board.put(6, 2, Piece(PieceType.SOLDIER))
        board.put(6, 4, Piece(PieceType.SOLDIER))
        board.put(6, 6, Piece(PieceType.SOLDIER))
        board.put(6, 8, Piece(PieceType.SOLDIER))

        board.put(7, 1, Piece(PieceType.CANNON))
        board.put(7, 7, Piece(PieceType.CANNON))

        board.put(8, 4, Piece(PieceType.GENERAL))

        board.put(9, 3, Piece(PieceType.GUARD))
        board.put(9, 5, Piece(PieceType.GUARD))

        board.put(9, 0, Piece(PieceType.CHARIOT))
        board.put(9, 8, Piece(PieceType.CHARIOT))

        if formation == Formation.OUTER_ELEPHANT:
            board.put(9, 1, Piece(PieceType.ELEPHANT))
            board.put(9, 2, Piece(PieceType.HORSE))
            board.put(9, 6, Piece(PieceType.HORSE))
            board.put(9, 7, Piece(PieceType.ELEPHANT))

        elif formation == Formation.LEFT_ELEPHANT:
            board.put(9, 1, Piece(PieceType.ELEPHANT))
            board.put(9, 2, Piece(PieceType.HORSE))
            board.put(9, 6, Piece(PieceType.ELEPHANT))
            board.put(9, 7, Piece(PieceType.HORSE))

        elif formation == Formation.RIGHT_ELEPHANT:
            board.put(9, 1, Piece(PieceType.HORSE))
            board.put(9, 2, Piece(PieceType.ELEPHANT))
            board.put(9, 6, Piece(PieceType.HORSE))
            board.put(9, 7, Piece(PieceType.ELEPHANT))

        elif formation == Formation.INNER_ELEPHANT:
            board.put(9, 1, Piece(PieceType.HORSE))
            board.put(9, 2, Piece(PieceType.ELEPHANT))
            board.put(9, 6, Piece(PieceType.ELEPHANT))
            board.put(9, 7, Piece(PieceType.HORSE))
        return board



from . import constants
from .base.board import Board
from .base.camp import Camp
from .base.formation import Formation
from .base.location import Location
from .base.move import MoveSet
from .base.piece import Piece, PieceType
from .game.janggi_game import JanggiGame
from .game.game_log import GameLog
from .ui.game_window import GameWindow
from .ui.replay_viewer import ReplayViewer
from .proto import log_pb2
from .utils import generate_random_game, play, replay


import pygame

GREEN = (0, 255, 0)

class BoardMarker:
    def __init__(self, screen, x: int, y: int):
        self.screen = screen
        self.x = x
        self.y = y

    def draw(self):
        pygame.draw.circle(self.screen, GREEN, (self.x, self.y), 10)


import logging
import pygame
from pygame.locals import *

from ..game.game_log import GameLog
from .game_window import GameWindow

class ReplayViewer:
    """Display replay of a single game using GameWindow."""

    def __init__(self, game_log: GameLog):
        self.window = GameWindow(game_log.board_log[0])
        self.game_log = game_log

    def run(self):
        self.window.render()
        pygame.event.clear()
        while True:
            event = pygame.event.wait()
            if event.type == QUIT:
                self.window.close()
            if event.type == KEYDOWN:
                try:
                    if event.key == K_ESCAPE:
                        self.window.close()
                        return
                    elif event.key == K_RIGHT:
                        board = self.game_log.next()
                        self.window.switch_board(board)
                        self.window.render()
                    elif event.key == K_LEFT:
                        board = self.game_log.prev()
                        self.window.switch_board(board)
                        self.window.render()
                except StopIteration:
                    logging.info("Reached end of the game log.")






import logging
import random

from .base.camp import Camp
from .base.formation import Formation
from .game.janggi_game import JanggiGame
from .game.game_log import GameLog
from .ui.game_player import GamePlayer
from .ui.replay_viewer import ReplayViewer
from .proto import log_pb2

logging.basicConfig()
logging.root.setLevel(logging.DEBUG)

def replay(filepath: str):
    """
    Replay a game by parsing the log file at the given path.

    Args:
        filepath (str): Path of the proto-serialized log file.
    """
    log_file = open(filepath, "rb")
    log_proto = log_pb2.Log()
    log_proto.ParseFromString(log_file.read())
    game_log = GameLog.from_proto(log_proto)
    game_log.generate_board_log()
    replay_viewer = ReplayViewer(game_log)
    replay_viewer.run()

def play(game: JanggiGame):
    """
    Play a game by running GamePlayer.

    Args:
        game (JanggiGame): Pre-initialized game to play.
    """
    player = GamePlayer(game)
    player.run()

def generate_random_game():
    """Generate a random Janggi game."""
    camp = Camp(random.choice([-1, 1]))
    cho_formation = Formation(random.randint(1, 4))
    han_formation = Formation(random.randint(1, 4))
    return JanggiGame(camp, cho_formation, han_formation)


from typing import List, Tuple

from ..constants import MIN_ROW, MAX_ROW, MIN_COL, MAX_COL
from .location import Location
from .camp import Camp
from .piece import PieceType


class MoveSet:
    """
    MoveSet is a list of a piece's moves in order to complete a single action.
    MoveSet ~ [(dr1, dc1), (dr2, dc2), (dr3, dc3), ...].
    If a player makes an action to move a Chariot piece 3 tiles up,
    its move set will be [(-1,0), (-1,0), (-1,0)].
    """
    def __init__(self, moves: List[Tuple[int, int]]):
        """Initialize MoveSet with the given list of moves."""
        self.moves = moves

    def __str__(self) -> str:
        """Return string representation of MoveSet."""
        return str(self.moves)

    def get_dest(self, board: "Board", origin: Location, player: Camp) -> Location:
        """
        Get destination location of piece at given origin location for the given player.

        Args:
            board (Board): Current board being played.
            origin (Location): Original location of the piece being played.
            player (Camp): Current player to make an action.

        Returns:
            Location: Destination of the piece if it's played through self.moves from origin.
        """
        if self.is_valid(board, origin, player):
            row, col = (origin.row, origin.col)
            for dr, dc in self.moves:
                row += dr
                col += dc
            return Location(row, col)
        else:
            return None

    def is_valid(self, board: "Board", origin: Location, player: Camp) -> bool:
        """
        Check validity of a move set for piece at the given origin for the given player.

        Args:
            board (Board): Current board being played.
            origin (Location): Original location of the piece being played.
            player (Camp): Current player to make an action.

        Returns:
            bool: True if move set is valid; False otherwise
        """
        def _is_out_of_bound(row: int, col: int):
            return (row < MIN_ROW or row > MAX_ROW or
                    col < MIN_COL or col > MAX_COL)
        origin_piece = board.get(origin.row, origin.col)
        num_hurdles = 1 if origin_piece.piece_type == PieceType.CANNON else 0
        row, col = (origin.row, origin.col)
        for i in range(len(self.moves)):
            (dr, dc) = self.moves[i]
            row += dr
            col += dc
            if _is_out_of_bound(row, col):
                return False

            piece = board.get(row, col)

            if i == len(self.moves)-1:
                # invalidate if landing on an ally piece
                if piece and piece.camp == player:
                    return False
                # invalidate if some hurdles are left for a cannon
                if origin_piece.piece_type == PieceType.CANNON and num_hurdles > 0:
                    return False
                # invalidate if cannon is landing on another cannon
                if (origin_piece.piece_type == PieceType.CANNON and 
                    (piece and piece.piece_type == PieceType.CANNON)):
                    return False

            elif piece:
                if origin_piece.piece_type == PieceType.CANNON:
                    # cannon cannot ever pass cannon
                    if piece.piece_type == PieceType.CANNON:
                        return False
                    # decrement number of hurdles left when passing a piece
                    num_hurdles -= 1
                    if num_hurdles < 0:
                        return False
                else:
                    # invalidate if there's a blocking piece before move set is complete
                    return False
        return True



from enum import Enum
from termcolor import colored
from typing import List

from ..constants import (
    MIN_ROW, MAX_ROW, MIN_COL, MAX_COL,
    CASTLE_MIN_COL, CASTLE_MAX_COL,
    CASTLE_TOP_MIN_ROW, CASTLE_TOP_MAX_ROW,
    CASTLE_BOT_MIN_ROW, CASTLE_BOT_MAX_ROW,
    CASTLE_TOP_SOLDIER_EXCEPTION_LEFT,
    CASTLE_TOP_SOLDIER_EXCEPTION_RIGHT,
    CASTLE_BOT_SOLDIER_EXCEPTION_LEFT,
    CASTLE_BOT_SOLDIER_EXCEPTION_RIGHT,
)
from .camp import Camp
from .location import Location


class PieceType(Enum):
    """Enum class representing a piece's type."""
    GENERAL = 1
    GUARD = 2
    HORSE = 3
    ELEPHANT = 4
    CHARIOT = 5
    CANNON = 6
    SOLDIER = 7

PIECE_VALUE = {
    PieceType.GUARD: 3,
    PieceType.HORSE: 5,
    PieceType.ELEPHANT: 3,
    PieceType.CHARIOT: 13,
    PieceType.CANNON: 7,
    PieceType.SOLDIER: 2,
    PieceType.GENERAL: 0,
}

from .move import MoveSet # Imported here to avoid an import loop.
class Piece:
    """
    Piece class represents a single piece on the game board.
    The class is capable of getting move sets based on its piece type.
    """
    def __init__(self, piece_type: PieceType):
        """
        Initialize Piece class by setting the given piece type.
        Camp attribute is later initialized by Board.mark_camp.
        """
        self.piece_type = piece_type
        self.camp = None

    def __float__(self) -> float:
        """
        Get unique float value for the piece's type.
        For all camp cho's pieces, this returns +(self.piece_type.value)
        For all camp han's pieces, this returns -(self.piece_type.value)

        Raises:
            Exception: if a piece does not have a camp assigned yet.

        Returns:
            float: uniqued float value for the piece's type.
        """
        if self.camp == None:
            raise Exception(
                f"{self.piece_type} does not have a camp assigned.")

        if self.camp == Camp.HAN:
            return float(self.piece_type.value) * -1.0
        else:
            return float(self.piece_type.value)

    @property
    def value(self) -> int:
        """
        Return piece's value based on its piece type. Each piece in Janggi has a 
        value assigned, and the values contribute to the players' current scores.

        Returns:
            int: piece's value based on type.
        """
        return PIECE_VALUE[self.piece_type]

    def __str__(self) -> str:
        """
        Return string representation of piece while applying different colors for each camp.
        """
        print_str = self._piece_to_chinese_character()
        if self.camp == Camp.CHO:
            print_str = colored(print_str, 'green')
        elif self.camp == Camp.HAN:
            print_str = colored(print_str, 'red')
        return print_str

    def _piece_to_chinese_character(self) -> str:
        """
        Return a Chinese character for the piece based on its type and camp.

        Returns:
            str: a single Chinese character for the piece.
        """
        piece_name_dict = {
            PieceType.GUARD: "士",
            PieceType.HORSE: "馬",
            PieceType.ELEPHANT: "象",
            PieceType.CHARIOT: "車",
            PieceType.CANNON: "包",
        }

        if self.piece_type in piece_name_dict:
            return piece_name_dict[self.piece_type]
        else:
            if self.piece_type == PieceType.GENERAL:
                if self.camp == Camp.CHO:
                    return "楚"
                else:
                    return "漢"
            elif self.piece_type == PieceType.SOLDIER:
                if self.camp == Camp.CHO:
                    return "卒"
                else:
                    return "兵"

    def get_soldier_move_sets(self, origin: Location, is_player: bool) -> List[MoveSet]:
        """
        Get move sets for soldier pieces.
        The directions a soldier piece can take depends on which camp it belongs to.

        Args:
            is_player (bool): True if the piece belongs to the main player; False otherwise.

        Returns:
            List[MoveSet]: All move sets a soldier piece can make regardless of validity.
        """
        steps = [(0, -1), (0, 1)]
        steps += [(-1, 0)] if is_player else [(1, 0)]
        # soldiers can move diagonally in enemy's castle
        if is_player:
            if tuple(origin) == CASTLE_TOP_SOLDIER_EXCEPTION_LEFT:
                steps += [(-1, 1)]
            elif tuple(origin) == CASTLE_TOP_SOLDIER_EXCEPTION_RIGHT:
                steps += [(-1, -1)]
        else:
            if tuple(origin) == CASTLE_BOT_SOLDIER_EXCEPTION_LEFT:
                steps += [(1, 1)]
            elif tuple(origin) == CASTLE_BOT_SOLDIER_EXCEPTION_RIGHT:
                steps += [(1, -1)]
        return [MoveSet([(dr, dc)]) for (dr, dc) in steps]

    def get_castle_move_sets(self, origin: Location, is_player: bool) -> List[MoveSet]:
        """
        Get move sets for castle pieces (generals and guards).
        The directions a soldier piece can take depends on which camp it belongs to.

        Args:
            origin (Location): Original location of the piece.
            is_player (bool): True if the piece belongs to the main player; False otherwise.

        Returns:
            List[MoveSet]: All move sets a castle piece can make regardless of validity.
        """
        def _is_out_of_bound(row: int, col: int):
            return (col < CASTLE_MIN_COL or col > CASTLE_MAX_COL or
                (is_player and (row < CASTLE_BOT_MIN_ROW or row > CASTLE_BOT_MAX_ROW)) or
                (not is_player and (row < CASTLE_TOP_MIN_ROW or row > CASTLE_TOP_MAX_ROW)))
        steps = [(i, j) for i in range(-1, 2) for j in range(-1, 2) if i != 0 or j != 0]
        steps = [(i, j) for (i, j) in steps if not _is_out_of_bound(origin.row+i, origin.col+j)]
        return [MoveSet([(dr, dc)]) for (dr, dc) in steps]

    def get_jumpy_move_sets(self) -> List[MoveSet]:
        """
        Get move sets for jumpy pieces (horses and elephants).

        Returns:
            List[MoveSet]: All move sets a jumpy piece can make regardless of validity.
        """
        move_sets = []
        if self.piece_type == PieceType.HORSE:
            move_sets.append(MoveSet([(0, 1), (-1, 1)]))
            move_sets.append(MoveSet([(0, 1), (1, 1)]))
            move_sets.append(MoveSet([(1, 0), (1, 1)]))
            move_sets.append(MoveSet([(1, 0), (1, -1)]))
            move_sets.append(MoveSet([(0, -1), (-1, -1)]))
            move_sets.append(MoveSet([(0, -1), (1, -1)]))
            move_sets.append(MoveSet([(-1, 0), (-1, 1)]))
            move_sets.append(MoveSet([(-1, 0), (-1, -1)]))
        elif self.piece_type == PieceType.ELEPHANT:
            move_sets.append(MoveSet([(0, 1), (-1, 1), (-1, 1)]))
            move_sets.append(MoveSet([(0, 1), (1, 1), (1, 1)]))
            move_sets.append(MoveSet([(1, 0), (1, 1), (1, 1)]))
            move_sets.append(MoveSet([(1, 0), (1, -1), (1, -1)]))
            move_sets.append(MoveSet([(0, -1), (-1, -1), (-1, -1)]))
            move_sets.append(MoveSet([(0, -1), (1, -1), (1, -1)]))
            move_sets.append(MoveSet([(-1, 0), (-1, 1), (-1, 1)]))
            move_sets.append(MoveSet([(-1, 0), (-1, -1), (-1, -1)]))
        return move_sets

    def get_straight_move_sets(self, origin: Location) -> List[MoveSet]:
        """
        Get move sets for straight pieces (chariots and cannons).

        Args:
            origin (Location): Original location of the piece.

        Returns:s
            List[MoveSet]: All move sets a straight piece can make regardless of validity.
        """
        def _is_out_of_bound(row: int, col: int):
            return (row < MIN_ROW or row > MAX_ROW or
                    col < MIN_COL or col > MAX_COL)

        def _get_move_sets_in_direction(origin: Location, dr: int, dc: int):
            (row, col) = (origin.row, origin.col)
            steps = []
            move_sets = []
            while not _is_out_of_bound(row+dr, col+dc):
                row += dr
                col += dc
                steps.append((dr, dc))
                move_sets.append(MoveSet(steps.copy()))
            return move_sets

        move_sets = []
        for (dr, dc) in [(0, -1), (0, 1), (-1, 0), (1, 0)]:
            move_sets += _get_move_sets_in_direction(origin, dr, dc)
        return move_sets



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="janggi",
    version="0.9.19",
    author="Sungho Cho",
    author_email="didog9595@gmail.com",
    description="Python library for Korean chess Janggi",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sungho-cho/pyjanggi",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=["janggi", "janggi/base", "janggi/game", "janggi/ui", "janggi/proto"],
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=["termcolor==1.1.0", "pygame==2.1.2"],
)



import logging
import os
import pkg_resources
import pygame
from typing import List, Optional

from ..base.board import Board
from ..base.camp import Camp
from ..base.location import Location
from ..base.piece import PieceType
from ..constants import MIN_ROW, MAX_ROW, MIN_COL, MAX_COL
from .marker import BoardMarker

WIDTH, HEIGHT = 500, 500
BOARD_WIDTH, BOARD_HEIGHT = 500, 500
PIECE_WIDTH, PIECE_HEIGHT = 40, 40
BOARD_Y, BOARD_X = 10, 10
ROW_GAP, COL_GAP = 50, 55
IMG_PATH = "images/"
BOARD_FILENAME = "board.png"

class GameWindow:
    """Class that renders board display using pygame."""
    def __init__(self, board: Optional[Board] = None):
        pygame.init()
        pygame.display.init()
        pygame.display.set_caption("Janggi")
        self.display = pygame.display.set_mode((WIDTH, HEIGHT))
        self.board = board
        self.board_img = None
        self.piece_imgs = {Camp.CHO: {}, Camp.HAN: {}}
        self.board_markers = []

        self._initialize_board_image()
        self._initialize_piece_images()

    def render(self):
        self.display.blit(self.board_img, (0, 0))
        # Draw pieces
        if self.board:
            for row, col in self.board.get_piece_locations():
                piece = self.board.get(row,col)
                if piece.camp is None:
                    logging.warning(f"(GameWindow) Camp is not assigned for piece {piece.piece_type}")
                    continue
                piece_img = self.piece_imgs[piece.camp][piece.piece_type]
                x, y = self.get_board_xy(row, col)
                self.display.blit(
                    piece_img, 
                    (
                        x - PIECE_WIDTH // 2, 
                        y - PIECE_HEIGHT // 2
                    )
                )
        # Draw markers
        for marker in self.board_markers:
            marker.draw()

        pygame.event.pump()
        pygame.display.update()

    def close(self):
        pygame.quit()

    def switch_board(self, board: Board):
        self.board = board
        pygame.event.pump()

    def get_board_xy(self, row: int, col: int):
        return (
            BOARD_X + COL_GAP * col + PIECE_WIDTH // 2, 
            BOARD_Y + ROW_GAP * row + PIECE_HEIGHT // 2
        )
    
    def _initialize_board_image(self):
        board_path = os.path.join(IMG_PATH, BOARD_FILENAME)
        board_file = pkg_resources.resource_filename(__name__, board_path)
        board_img = pygame.image.load(board_file)
        self.board_img = pygame.transform.scale(board_img, (BOARD_WIDTH, BOARD_HEIGHT))

    def _initialize_piece_images(self):
        for camp in Camp:
            if camp == Camp.UNDECIDED:
                continue
            for piece_type in PieceType:
                piece_path = self._get_image_path(camp, piece_type)
                piece_file = pkg_resources.resource_filename(__name__, piece_path)
                piece_img = pygame.image.load(piece_file)
                self.piece_imgs[camp][piece_type] = pygame.transform.scale(
                    piece_img, (PIECE_WIDTH, PIECE_HEIGHT))

    def _get_image_path(self, camp: Camp, piece_type: PieceType):
        filename = f"{camp.name.lower()}_{piece_type.name.lower()}.png"
        return os.path.join(IMG_PATH, filename)



from __future__ import annotations
from enum import IntEnum
from ..proto import log_pb2

class Camp(IntEnum):
    """
    Enum that represents two sides of the game of Janggi.
    It's just like black and white in chess except it's called cho and han.
    """
    CHO = 1
    HAN = -1
    UNDECIDED = 0

    @property
    def opponent(self) -> Camp:
        """
        Return opponent's enum. (Camp.CHO for Camp.HAN / Camp.HAN for Camp.CHO).

        Returns:
            Camp: Opponent's enum instance.
        """
        return Camp(self * -1)

    @classmethod
    def from_proto(cls, camp_proto: log_pb2.Camp) -> Camp:
        """Convert from proto Camp enum."""
        if camp_proto == log_pb2.Camp.CHO:
            return cls.CHO
        elif camp_proto == log_pb2.Camp.HAN:
            return cls.HAN
        else:
            return cls.UNDECIDED

    def to_proto(self) -> log_pb2.Camp:
        """Convert to proto Camp enum."""
        if self == Camp.CHO:
            return log_pb2.Camp.CHO
        elif self == Camp.HAN:
            return log_pb2.Camp.HAN
        else:
            raise Exception(f"Cannot convert Camp enum {self} to proto.")
        


from __future__ import annotations
from enum import IntEnum
from ..proto import log_pb2


class Formation(IntEnum):
    """
    Enum class that represents 4 different formations of the game of Janggi.
    See https://en.wikipedia.org/wiki/Janggi#Setting_up for details on formations.
    In Korean terminology,
    the inner elephant setup is called Won-Ang-Ma,
    the outer elephant setup is called Yang-Gwi-Ma,
    the left elephant and right elephant setup are called Gwi-Ma.
    """
    INNER_ELEPHANT = 1
    OUTER_ELEPHANT = 2
    LEFT_ELEPHANT = 3
    RIGHT_ELEPHANT = 4
    UNDECIDED = 5

    @classmethod
    def from_proto(cls, formation_proto: log_pb2.Formation) -> Formation:
        """Convert from proto Formation enum."""
        if formation_proto == log_pb2.Formation.INNER_ELEPHANT:
            return cls.INNER_ELEPHANT
        elif formation_proto == log_pb2.Formation.OUTER_ELEPHANT:
            return cls.OUTER_ELEPHANT
        if formation_proto == log_pb2.Formation.LEFT_ELEPHANT:
            return cls.LEFT_ELEPHANT
        elif formation_proto == log_pb2.Formation.RIGHT_ELEPHANT:
            return cls.RIGHT_ELEPHANT
        else:
            return cls.UNDECIDED

    def to_proto(self) -> log_pb2.Formation:
        """Convert to proto Formation enum."""
        if self == Formation.INNER_ELEPHANT:
            return log_pb2.Formation.INNER_ELEPHANT
        elif self == Formation.OUTER_ELEPHANT:
            return log_pb2.Formation.OUTER_ELEPHANT
        elif self == Formation.LEFT_ELEPHANT:
            return log_pb2.Formation.LEFT_ELEPHANT
        elif self == Formation.RIGHT_ELEPHANT:
            return log_pb2.Formation.RIGHT_ELEPHANT
        else:
            raise Exception(f"Cannot convert Formation enum {self} to proto.")